#include "stdafx.h"
#include "windows.h"
#include "math.h"
#include "cuda.h"
#include "cufft.h"
#include <iostream>
using namespace std;
#pragma comment(lib, "cufft.lib")
#pragma comment(lib, "cuda.lib")

//Check function return value
inline void CheckStatus(const char *szInfo, const int nStatus)
{
	if(0 != nStatus)
	{
		cout << szInfo << " fail! Error Code: " << nStatus << endl;
		exit(-1);
	}
}

#define NX 16384 //will be fail
//#define NX 8192  //will be successfull
#define BATCH 128

int _tmain(int argc, _TCHAR* argv[])
{
	CUdevice dev;
	CUcontext ctx;


	CUresult status = cuInit(0);
	CheckStatus("cuInit", status);


	int nGpuCount = 0;
	status = cuDeviceGetCount(&nGpuCount);
	CheckStatus("cuDeviceGetCount", status);
	if(0 == nGpuCount)
	{
		cout << "There is no GPU available!" << endl;
		exit(-1);
	}


	int nOrdinal = 0;
	int nMPNum = 0; 
	int nTmpMPNum = 0;
	CUdevice devtmp;
	//Find one device with more processors.
	for(int i = 0; i < nGpuCount; i++)
	{
		status = cuDeviceGet(&devtmp, i);
		CheckStatus("cuDeviceGet", status);

		status = cuDeviceGetAttribute(&nTmpMPNum, CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT, devtmp);
		CheckStatus("cuDeviceGetAttribute", status);

		if(nTmpMPNum > nMPNum)
		{
			nOrdinal = i;
			nMPNum = nTmpMPNum;
			printf("Multi processor num: %d\n", nMPNum);//for debug
		}
	}


	status = cuDeviceGet(&dev, nOrdinal);
	CheckStatus("cuDeviceGet", status);


	status = cuCtxCreate(&ctx, CU_CTX_SCHED_SPIN, dev);
	CheckStatus("cuCtxCreate", status);


	cufftHandle plan;
	cufftResult result = CUFFT_SUCCESS;
	CUresult ret = CUDA_SUCCESS;
	int nx = NX;
	int batch = BATCH;


	result = cufftPlan1d(&plan, nx, CUFFT_C2C, batch);
	CheckStatus("cufftPlan1d", result);


	cufftComplex *idata_h = (cufftComplex*)malloc((nx) * batch * sizeof(cufftComplex));
	if(NULL == idata_h)
	{
		cout << "malloc fail! " << endl;
		exit(-1);
	}


	CUdeviceptr idata_d;
	ret = cuMemAlloc(&idata_d, (nx) * batch * sizeof(cufftComplex));
	CheckStatus("cuMemAlloc", ret);


	//Initialize input data
	srand(GetTickCount());
	for(int i = 0; i < (nx)*batch * 2; i++)
	{
		((float*)idata_h)[i] = (double)rand() / (RAND_MAX + 1);
	}


	ret = cuMemcpyHtoD(idata_d, idata_h, (nx)*batch*sizeof(cufftComplex));
	CheckStatus("cuMemcpyHtoD", ret);


	result = cufftExecC2C(plan, (cufftComplex *)idata_d, (cufftComplex *)idata_d, CUFFT_FORWARD);
	CheckStatus("CUFFT_FORWARD", ret);


	ret = cuCtxSynchronize();//LOOK HERE: return value will be CUDA_ERROR_LAUNCH_FAILED when nx=16384
	CheckStatus("First cuCtxSynchronize", ret);


	result = cufftExecC2C(plan, (cufftComplex *)idata_d, (cufftComplex *)idata_d, CUFFT_INVERSE);
	CheckStatus("CUFFT_INVERSE", result);
	

	ret = cuCtxSynchronize();//LOOK HERE: return value will be CUDA_ERROR_LAUNCH_FAILED when nx=16384
	CheckStatus("Second cuCtxSynchronize", ret);

	
	result = cufftDestroy(plan);
	CheckStatus("cufftDestroy", result);


	ret = cuMemFree(idata_d);
	CheckStatus("cuMemFree", ret);


	ret = cuCtxDetach(ctx);
	CheckStatus("cuCtxDetach", ret);

	ret = cuCtxDestroy(ctx);
	CheckStatus("cuCtxDestroy", ret);
	

	free(idata_h);

	cout << "Success! " << endl << "Press any key to continue!" << endl;

	getchar();

	return 0;
}
